#include <cstdio>
#include <vector>
#include <map>

#include "level.hh"
#include "defs.hh"
#include "rom.hh"
#include "objects.hh"

void Level::WritePalette(unsigned addr, unsigned addr_afterdoor) const
{
    ROM.WriteBytes(addr,           16*3, Palette);
    
    if(LevelNumber < 6)
    {
        ROM.WriteBytes(addr_afterdoor, 16, AfterDoorPalette);
    }
}

void Level::WriteActives(unsigned first_screen, unsigned addr, unsigned end_addr) const
{
    unsigned req_size = end_addr - addr;
    
    std::vector<unsigned char> activebuf;
    
    activebuf.push_back(0);
    
    unsigned n=0;
    for(unsigned a=0; a<MaxActives; ++a)
    {
        if(actives[a].screen == 255) continue;
        activebuf.push_back(actives[a].type);
        activebuf.push_back(actives[a].screen + first_screen);
        activebuf.push_back(actives[a].x1);
        activebuf.push_back(actives[a].y1);
        activebuf.push_back(actives[a].x2);
        activebuf.push_back(actives[a].y2);
        ++n;
    }
    activebuf[0] = n+1;
    
    activebuf.push_back(0xFF);
    do { activebuf.push_back(0xFF); }
    while(activebuf.size() < req_size);
    
    if(activebuf.size() > req_size)
    {
        fprintf(stderr, "Error: Too many actives!\n");
    }
    
    ROM.WriteBytes(addr, req_size, &activebuf[0]);
}

void Level::WriteEnemies
    (unsigned first_screen,
     unsigned ptraddr, unsigned& addr, unsigned end_addr) const
{
    if(LevelNumber >= 10) return;
    
    ROM.WritePointer(ptraddr, addr);
    
    std::vector<unsigned char> enemybuf;

    for(unsigned a=0; a<MaxEnemies; ++a)
    {
        if(enemies[a].screen == 255)
        {
            enemybuf.push_back(0xFF);
            break;
        }
        else
        {
            unsigned screen = enemies[a].screen;
            if(screen != 128) screen += first_screen;
            enemybuf.push_back(screen);
            enemybuf.push_back(enemies[a].x);
            enemybuf.push_back(enemies[a].y);
            enemybuf.push_back(enemies[a].type);
        }
    }
    
    unsigned max_size = end_addr - addr;
    if(enemybuf.size() > max_size)
    {
        fprintf(stderr, "Error: Too many enemies in the game (%u)!\n", enemybuf.size());
    }
    
    ROM.WriteBytes(addr, enemybuf.size(), &enemybuf[0]);
    addr += enemybuf.size();
}

void Level::ReadPalette(unsigned addr, unsigned addr_afterdoor)
{
    ROM.ReadData(addr,           16*3, Palette);
    ROM.ReadData(addr_afterdoor, 16, AfterDoorPalette);
}

void Level::ReadActives(unsigned first_screen, unsigned addr)
{
    unsigned n = ROM.Read(addr++);
    if(n > MaxActives)n=MaxActives;
    for(unsigned a=0; a<MaxActives; ++a) actives[a].Delete();
    for(unsigned a=0; a<n; ++a)
    {
        unsigned char Buf[6];
        ROM.ReadData(addr + a*6, 6, Buf);
        
        if(Buf[1] == 255) break;
        
        //fprintf(stderr, "Reading active from %X\n", addr);
        actives[a].type   = Buf[0];
        actives[a].screen = Buf[1] - first_screen;
        actives[a].x1     = Buf[2];
        actives[a].y1     = Buf[3];
        actives[a].x2     = Buf[4];
        actives[a].y2     = Buf[5];
    }
}

void Level::ReadEnemies(unsigned first_screen, unsigned ptraddr)
{
    if(LevelNumber >= 10) return;
    
    unsigned data_addr = ROM.ReadPointer(ptraddr);
    
    //printf("// Enemies %04X -> %06X\n", ptraddr, data_addr);
    for(unsigned a=0; a<MaxEnemies; ++a) enemies[a].Delete();
    for(unsigned a=0; a<MaxEnemies; ++a)
    {
        unsigned char Buf[4];
        ROM.ReadData(data_addr + a*4, 4, Buf);

        if(Buf[0] == 255) break;
        /*
        printf("// enemy %u @ %X: %02X %02X %02X %02X\n",
          a, data_addr+a*4,
          Buf[0], Buf[1], Buf[2], Buf[3]);
        */
        enemies[a].screen = Buf[0];
        if(enemies[a].screen != 128) enemies[a].screen -= first_screen;
        enemies[a].x      = Buf[1];
        enemies[a].y      = Buf[2];
        enemies[a].type   = Buf[3];
    }
}

void Level::DeleteGutsBlockActives()
{
    /* Gutsblocks. */
    for(unsigned s=0; s<MaxRooms; ++s)
    {
        Screen& sc = rooms[s];
        if(sc.cont_dir == NoRooms) break;
        
        for(unsigned y=0; y<16; ++y)
            for(unsigned x=0; x<16; ++x)
              if(sc.ascii[y*32+x] == Ascii_Guts)
                  for(unsigned a=0; a<MaxActives; ++a)
                  {
                      Active& act = actives[a];
                      if(act.screen == s
                      && act.type == 2
                      && act.x1 == x*16
                      && act.y1 == y*16
                      && act.x2 == (((x+2)*16) & 0xFF)
                      && act.y2 == (((y+2)*16) & 0xFF))
                      {
                          act.Delete();
                      }
                  }
    }
}

void Level::MakeAsciiActives()
{
    unsigned buzzcounter[MaxRooms] = {0};
    
    // Buzzblocks are ordered last-to-first.
    // Thus we have to walk the structure backwards.
    for(unsigned a=MaxActives; a-- > 0; )
    {
        Active& act = actives[a];
        if(act.screen == 255) continue;
/**/
        if(act.type == 1 && act.y2 <= act.y1)
        {
            // flying fire
            unsigned x1 = act.x1; x1 /= 16;
            unsigned x2 = act.x2; x2 /= 16;
            unsigned y = act.y1; y /= 16;
            for(unsigned x=x1; x<x2; ++x)
                rooms[act.screen].ascii[y*32+16+x] = Ascii_ObjFire;
            
            act.Delete();
        }
/**/
        if(act.type == 1 && act.y2 > act.y1)
        {
            // leftside shutter
            unsigned x1 = act.x1; x1 /= 16;
            unsigned y1 = act.y1; y1 /= 16;
            unsigned x2 = act.x2; if(x2==0) x2=0x100;  x2 /= 16;
            unsigned y2 = act.y2; if(y2==0) y2=0x100;  y2 /= 16;
            
            /*
            for(unsigned y=y1; y<y2; ++y)
                for(unsigned x=x1; x<x2; ++x)
                    rooms[act.screen].ascii[y*32+16+x] = Ascii_ObjShutterIn;
            */
            
            act.Delete();
        }
        if(act.type == 3)
        {
            // rightside shutter
            unsigned x1 = act.x1; x1 /= 16;
            unsigned y1 = act.y1; y1 /= 16;
            unsigned x2 = act.x2; if(x2==0) x2=0x100;  x2 /= 16;
            unsigned y2 = act.y2; if(y2==0) y2=0x100;  y2 /= 16;
            
            for(unsigned y=y1; y<y2; ++y)
                for(unsigned x=x1; x<x2; ++x)
                    rooms[act.screen].ascii[y*32+16+x] = Ascii_ObjShutterOut;
            
            act.Delete();
        }

        if(act.type == 0 && false)
        {
            // buzz (appearing and disappearing blocks)
            unsigned x1 = act.x1; x1 /= 16;
            unsigned y1 = act.y1; y1 /= 16;
            unsigned x2 = act.x2; if(x2==0) x2=0x100;  x2 /= 16;
            unsigned y2 = act.y2; if(y2==0) y2=0x100;  y2 /= 16;
            
            for(unsigned y=y1; y<y2; ++y)
                for(unsigned x=x1; x<x2; ++x)
                {
                    unsigned& count = buzzcounter[act.screen];
                    if(count > 9)
                    {
                        fprintf(stderr, "Too many buzzblocks in one screen!\n");
                        break;
                    }
                    rooms[act.screen].ascii[y*32+16+x] = "0123456789"[count++];
                }
            
            act.Delete();
        }
    }
}

void Level::MakeAsciiEnemies()
{
    for(unsigned a=0; a<MaxEnemies; ++a)
    {
        Enemy& e = enemies[a];
        if(e.screen == 255) continue;
        
        char ch = 0;
        if(e.type < NumAsciiObjects)
            ch = AsciiObjects[e.type].ascii;
        
        if(ch != 0)
        {
            unsigned x = e.x / 16;
            unsigned y = e.y / 16;
            
            //printf("[%u](@%u %u,%u)\n", a, e.screen, x,y);
            
            char& target = rooms[e.screen].ascii[y*32+16+x];
            if(target == ' '
            || (target == Ascii_ObjFire && ch == Ascii_ObjFire)
              )
            {
                target = ch;
                e.Delete();
            }
        }
    }
}

void Level::CreateGutsBlockActives()
{
    for(unsigned s=0; s<MaxRooms; ++s)
    {
        Screen& sc = rooms[s];
        if(sc.cont_dir == NoRooms) break;
        
        char ignore[16*16] = {0};
        for(unsigned y=0; y<16; ++y)
            for(unsigned x=0; x<16; ++x)
            {
                if(ignore[y*16+x]) continue;
                if(sc.ascii[y*32+x] == Ascii_Guts)
                {
                    AddActive(Active(2, s, x*16,y*16, (x+2)*16, (y+2)*16));
                    ignore[y*16+x+1]=1;
                    ignore[(y+1)*16+x]=1;
                    ignore[(y+1)*16+x+1]=1;
                }
            }
    }
}

void Level::ReadAsciiActives()
{
    unsigned shuttercount = 0;
    
    if(LevelNumber < 10)
    {
        fprintf(stderr, "Writing FF at %X (shut1)\n",
            Addr_Door1ScreenTable + LevelNumber);
    
        fprintf(stderr, "Writing FF at %X (shut2)\n",
            Addr_Door2ScreenTable + LevelNumber);

        ROM.Write(Addr_Door1ScreenTable + LevelNumber, 0xFF);
        ROM.Write(Addr_Door2ScreenTable + LevelNumber, 0xFF);
    }
    
    for(unsigned s=0; s<MaxRooms; ++s)
    {
        Screen& sc = rooms[s];
        if(sc.cont_dir == NoRooms) break;
        
        char ignore[16*16] = {0};
        for(unsigned y=0; y<16; ++y)
            for(unsigned x=0; x<16; ++x)
            {
                if(ignore[y*16+x]) continue;
                char c = sc.ascii[y*32+16+x];
                switch(c)
                {
                    case Ascii_ObjShutterIn:
                    case Ascii_ObjShutterOut:
                    {
                        unsigned y1=y;
                        unsigned y2=y;
                        while(y2<16 && sc.ascii[y2*32+16+x]==c) ++y2;
                        unsigned x1=x;
                        unsigned x2=x;
                        while(x2<16 && sc.ascii[y1*32+16+x2]==c) ++x2;
                        
                        fprintf(stderr, "Creating shutter %c %u,%u,%u,%u\n",
                            c,x1,y1,x2,y2);
                        
                        int t = c==Ascii_ObjShutterIn ? 1 : 3;
                        AddActive(Active(t, s, x1*16,y1*16, x2*16,y2*16));

                        for(unsigned Y=y1; Y<y2; ++Y)
                            for(unsigned X=x1; X<x2; ++X)
                                ignore[Y*16+X]=1;
                        
                        if(c == Ascii_ObjShutterOut && LevelNumber < 10)
                        {
                            switch(shuttercount)
                            {
                                case 0:
                                    fprintf(stderr, "Writing %02X at %X (shut1)\n",
                                        s, Addr_Door1ScreenTable + LevelNumber);
                                    ROM.Write(Addr_Door1ScreenTable + LevelNumber, s);
                                    break;
                                case 1:
                                default:
                                    fprintf(stderr, "Writing %02X at %X (shut2)\n",
                                        s, Addr_Door2ScreenTable + LevelNumber);
                                    ROM.Write(Addr_Door2ScreenTable + LevelNumber, s);
                                    break;
                            }
                            ++shuttercount;
                        }
                        
                        break;
                    }
                    case Ascii_ObjFire:
                    {
                        AddActive(Active(1, s, x*16, y*16, (x+1)*16, y*16));
                        break;
                    }
                    case Ascii_TeleportYcoord:
                    {
                        if(sc.cont_dir & LevelBegin)
                            ROM.Write(Addr_TeleportYcoordTable + LevelNumber, (y*16)+4);
                        else if(sc.cont_dir & RestartPointA)
                            ROM.Write(Addr_RestartTeleportYcoordTable1 + LevelNumber, (y*16)+4);
                        else if(sc.cont_dir & RestartPointB)
                            ROM.Write(Addr_RestartTeleportYcoordTable2 + LevelNumber, (y*16)+4);
                        else
                        {
                            fprintf(stderr,
                                "TeleportYcoord used at screen %u even though it's not a level begin or a restart point.\n",
                                s);
                        }
                        break;
                    }
                    case Ascii_Boss:
                    {
                        unsigned yoffs = 4; if(LevelNumber==5) yoffs=0; // gutsman is taller.
                        ROM.Write(Addr_BossXcoord + LevelNumber, x*16);
                        ROM.Write(Addr_BossYcoord + LevelNumber, y*16 + yoffs);
                        break;
                    }
                }
            }
    }
    if(shuttercount > 2)
    {
        fprintf(stderr,
            "Warning: Too many shutters (%u) in %s stage!\n", shuttercount, Name);
    }
}

void Level::ReadAsciiEnemies()
{
    for(unsigned s=0; s<MaxRooms; ++s)
    {
        Screen& sc = rooms[s];
        if(sc.cont_dir == NoRooms) break;
        
        for(unsigned y=0; y<16; ++y)
            for(unsigned x=0; x<16; ++x)
            {
                char c = sc.ascii[y*32+16+x];
                if(c==' ') continue;
                
                if(c == Ascii_ObjShutterIn
                || c == Ascii_ObjShutterOut
                || c == Ascii_TeleportYcoord
                || c == Ascii_Boss
                  ) continue;
                
                for(unsigned a=0; a<NumAsciiObjects; ++a)
                    if(c == AsciiObjects[a].ascii)
                    {
                        c=0;
                        AddEnemy(Enemy(s, x*16 + AsciiObjects[a].x,
                                          y*16 + AsciiObjects[a].y,
                                          a));
                        break;
                    }
                    
                if(c)
                {
                    fprintf(stderr, "Unrecognized object '%c'\n", c);
                }
            }
    }
}

void Level::ReadGFX(unsigned level_addr,unsigned index_addr,unsigned table_addr)
{
    std::memset(&levelgfx,  sizeof(levelgfx),0);
    std::memset(&monstergfx,sizeof(monstergfx),0);
    
    unsigned levcount = ROM.Read(level_addr);
    for(unsigned a=0; a<levcount && a<MaxLevelGfx; ++a)
    {
        unsigned char Buf[2] = {0,0};
        ROM.ReadData(level_addr + 1 + a*2, 2, Buf);
        
        levelgfx[a].index = Buf[0];
        levelgfx[a].count = Buf[1];
    }
    
    std::map<unsigned, unsigned> defined;
    
    for(unsigned a=0; a<MaxRooms; ++a)
    {
        Screen& screen = rooms[a];
        if(screen.cont_dir == NoRooms) break;
        
        const unsigned rom_index = ROM.Read(index_addr + a);
        
        std::map<unsigned,unsigned>::iterator i = defined.find(rom_index);
        
        if(i == defined.end())
        {
            unsigned new_index = defined.size();
            defined[rom_index] = new_index;
            
            if(new_index >= MaxMonsterGfx) new_index = MaxMonsterGfx-1;
            
            ROM.ReadData(table_addr + rom_index + 0, 5, monstergfx[new_index].index);
            ROM.ReadData(table_addr + rom_index + 5, 6, monstergfx[new_index].palette);
            
            screen.MonsterGraphicsIndex = new_index;
        }
        else
        {
            screen.MonsterGraphicsIndex = i->second;
        }
    }
}

void Level::WriteGFX(unsigned level_addr,unsigned index_addr,unsigned table_addr)
{
    unsigned levcount=0;
    for(unsigned a=0; a<MaxLevelGfx; ++a)
    {
        unsigned char Buf[2];
        
        Buf[0] = levelgfx[a].index;
        Buf[1] = levelgfx[a].count;
        
        if(Buf[1]) { ++levcount; }

        ROM.WriteBytes(level_addr + 1 + a*2, 2, Buf);
    }
    ROM.Write(level_addr, levcount);
    
    for(unsigned a=0; a<MaxRooms; ++a)
    {
        Screen& screen = rooms[a];
        if(screen.cont_dir == NoRooms) break;
        
        ROM.Write(index_addr+a, screen.MonsterGraphicsIndex);
    }
    for(unsigned a=0; a<MaxMonsterGfx; ++a)
    {
        ROM.WriteBytes(table_addr + a*11 + 0, 5, monstergfx[a].index);
        ROM.WriteBytes(table_addr + a*11 + 5, 6, monstergfx[a].palette);
    }
}

void LevelPairData::CombineGFX(bool Combine)
{
    for(unsigned pass=1; pass<=2; ++pass)
    {
        std::map<std::string, unsigned> defs;
        
        for(unsigned a=0; a<MaxMonsterGfx; ++a)
        {
            if(pass==1 || Combine)
            {
              std::string token1 = level1.monstergfx[a].to_string();
              if(token1[0] && token1[1]) {
              std::map<std::string, unsigned>::iterator i = defs.find(token1);
              if(i == defs.end()) { unsigned n = defs.size() * 11; defs[token1] = n; }
            } }
            if(pass==2 || Combine)
            {
              std::string token2 = level2.monstergfx[a].to_string();
              if(token2[0] && token2[1]) {
              std::map<std::string, unsigned>::iterator i = defs.find(token2);
              if(i == defs.end()) { unsigned n = defs.size() * 11; defs[token2] = n; }
            } }
        }
        
        for(unsigned a=0; a<MaxRooms; ++a)
        {
            if(pass==1 || Combine)
            { Screen& screen = level1.rooms[a];
              if(screen.cont_dir != NoRooms)
                screen.MonsterGraphicsIndex = defs[level1.monstergfx[screen.MonsterGraphicsIndex].to_string()];
            }
            
            if(pass==2 || Combine)
            { Screen& screen = level2.rooms[a];
              if(screen.cont_dir != NoRooms)
                screen.MonsterGraphicsIndex = defs[level2.monstergfx[screen.MonsterGraphicsIndex].to_string()];
            }
        }
        
        unsigned n=0;
        for(std::map<std::string, unsigned>::iterator
            i = defs.begin(); i != defs.end(); ++i)
        {
            if(n >= MaxMonsterGfx) break;
            
            const std::string& s = i->first;
            
            if(pass==1 || Combine)
            {
                std::memcpy(level1.monstergfx[i->second/11].index,   s.c_str()+0, 5);
                std::memcpy(level1.monstergfx[i->second/11].palette, s.c_str()+5, 6);
            }
            if(pass==2 || Combine)
            {
                std::memcpy(level2.monstergfx[i->second/11].index,   s.c_str()+0, 5);
                std::memcpy(level2.monstergfx[i->second/11].palette, s.c_str()+5, 6);
            }
            
            ++n;
        }
        if(n >= MaxMonsterGfx)
        {
            std::fprintf(stderr, "Error: Too many monster GFX tables in pair (%u)\n", n);
        }
        if(pass == 1 && Combine) break;
    }
}
